from typing import Mapping, Sequence

from rflx.ada import (
    ID,
    TRUE,
    Add,
    Aggregate,
    And,
    AndThen,
    Assignment,
    Call,
    CallStatement,
    CaseStatement,
    Constrained,
    Declaration,
    Div,
    Equal,
    Expr,
    ExpressionFunctionDeclaration,
    ForAllIn,
    FormalSubprogramDeclaration,
    FunctionSpecification,
    GenericProcedureInstantiation,
    Greater,
    GreaterEqual,
    If,
    In,
    Indexed,
    InOutParameter,
    Length,
    LessEqual,
    Mod,
    Mul,
    NamedAggregate,
    Not,
    NullStatement,
    Number,
    ObjectDeclaration,
    Old,
    OutParameter,
    Parameter,
    Postcondition,
    Precondition,
    ProcedureSpecification,
    Range,
    Selected,
    Size,
    Slice,
    Statement,
    Subprogram,
    SubprogramBody,
    SubprogramDeclaration,
    UnitPart,
    ValueRange,
    Variable,
)
from rflx.common import unique
from rflx.const import BUILTINS_PACKAGE
from rflx.model import FINAL, Enumeration, Field, Message, Opaque, Scalar, Type

from . import common, const


class SerializerGenerator:
    def __init__(self, prefix: str = "") -> None:
        self.prefix = prefix

    def insert_function(self, type_identifier: ID) -> Subprogram:
        return GenericProcedureInstantiation(
            "Insert",
            ProcedureSpecification(
                const.TYPES * "Insert",
                [
                    Parameter(["Val"], type_identifier),
                    InOutParameter(["Buffer"], const.TYPES_BYTES),
                    Parameter(["Offset"], const.TYPES_OFFSET),
                ],
            ),
            [common.prefixed_type_identifier(type_identifier, self.prefix)],
        )

    def create_internal_functions(
        self, message: Message, scalar_fields: Mapping[Field, Scalar]
    ) -> UnitPart:
        return UnitPart(
            [],
            [
                SubprogramBody(
                    ProcedureSpecification(
                        "Set_Field_Value",
                        [
                            InOutParameter(["Ctx"], "Context"),
                            Parameter(["Val"], "Field_Dependent_Value"),
                            OutParameter(["Fst", "Lst"], const.TYPES_BIT_INDEX),
                        ],
                    ),
                    [
                        *common.field_bit_location_declarations(Variable("Val.Fld")),
                        *common.field_byte_location_declarations(),
                        *unique(
                            self.insert_function(common.full_base_type_name(t))
                            for t in message.types.values()
                            if isinstance(t, Scalar)
                        ),
                    ],
                    [
                        Assignment("Fst", Variable("First")),
                        Assignment("Lst", Variable("Last")),
                        CaseStatement(
                            Variable("Val.Fld"),
                            [
                                (
                                    Variable(f.affixed_name),
                                    [
                                        CallStatement(
                                            "Insert",
                                            [
                                                Variable(f"Val.{f.name}_Value"),
                                                Slice(
                                                    Variable("Ctx.Buffer.all"),
                                                    Variable("Buffer_First"),
                                                    Variable("Buffer_Last"),
                                                ),
                                                Variable("Offset"),
                                            ],
                                        )
                                        if f in scalar_fields
                                        else NullStatement()
                                    ],
                                )
                                for f in message.all_fields
                            ],
                        ),
                    ],
                    [
                        Precondition(
                            AndThen(
                                Not(Constrained("Ctx")),
                                Call("Has_Buffer", [Variable("Ctx")]),
                                In(Variable("Val.Fld"), Range("Field")),
                                Call("Valid_Next", [Variable("Ctx"), Variable("Val.Fld")]),
                                common.sufficient_space_for_field_condition(Variable("Val.Fld")),
                                ForAllIn(
                                    "F",
                                    Range("Field"),
                                    If(
                                        [
                                            (
                                                Call(
                                                    "Structural_Valid",
                                                    [
                                                        Indexed(
                                                            Variable("Ctx.Cursors"),
                                                            Variable("F"),
                                                        )
                                                    ],
                                                ),
                                                LessEqual(
                                                    Selected(
                                                        Indexed(
                                                            Variable("Ctx.Cursors"),
                                                            Variable("F"),
                                                        ),
                                                        "Last",
                                                    ),
                                                    Call(
                                                        "Field_Last",
                                                        [Variable("Ctx"), Variable("Val.Fld")],
                                                    ),
                                                ),
                                            )
                                        ]
                                    ),
                                ),
                            )
                        ),
                        Postcondition(
                            And(
                                Call("Has_Buffer", [Variable("Ctx")]),
                                Equal(
                                    Variable("Fst"),
                                    Call("Field_First", [Variable("Ctx"), Variable("Val.Fld")]),
                                ),
                                Equal(
                                    Variable("Lst"),
                                    Call("Field_Last", [Variable("Ctx"), Variable("Val.Fld")]),
                                ),
                                GreaterEqual(Variable("Fst"), Variable("Ctx.First")),
                                LessEqual(Variable("Fst"), Add(Variable("Lst"), Number(1))),
                                LessEqual(Variable("Lst"), Variable("Ctx.Last")),
                                ForAllIn(
                                    "F",
                                    Range("Field"),
                                    If(
                                        [
                                            (
                                                Call(
                                                    "Structural_Valid",
                                                    [
                                                        Indexed(
                                                            Variable("Ctx.Cursors"),
                                                            Variable("F"),
                                                        )
                                                    ],
                                                ),
                                                LessEqual(
                                                    Selected(
                                                        Indexed(
                                                            Variable("Ctx.Cursors"),
                                                            Variable("F"),
                                                        ),
                                                        "Last",
                                                    ),
                                                    Variable("Lst"),
                                                ),
                                            )
                                        ]
                                    ),
                                ),
                                *const.CONTEXT_INVARIANT,
                                *[
                                    Equal(e, Old(e))
                                    for e in [
                                        Variable("Ctx.Cursors"),
                                    ]
                                ],
                            )
                        ),
                    ],
                )
            ]
            if scalar_fields
            else [],
        )

    def create_scalar_setter_procedures(
        self, message: Message, scalar_fields: Mapping[Field, Scalar]
    ) -> UnitPart:
        def specification(field: Field, field_type: Type) -> ProcedureSpecification:
            if field_type.package == BUILTINS_PACKAGE:
                type_identifier = ID(field_type.name)
            elif isinstance(field_type, Enumeration) and field_type.always_valid:
                type_identifier = common.prefixed_type_identifier(
                    common.full_enum_name(field_type), self.prefix
                )
            else:
                type_identifier = common.prefixed_type_identifier(
                    ID(field_type.identifier), self.prefix
                )

            return ProcedureSpecification(
                f"Set_{field.name}",
                [InOutParameter(["Ctx"], "Context"), Parameter(["Val"], type_identifier)],
            )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f, t),
                    [
                        Precondition(
                            AndThen(
                                *self.setter_preconditions(f),
                                Call(
                                    "Field_Condition",
                                    [
                                        Variable("Ctx"),
                                        Aggregate(
                                            Variable(f.affixed_name),
                                            Call("To_Base", [Variable("Val")]),
                                        ),
                                    ],
                                ),
                                Call("Valid", [Call("To_Base", [Variable("Val")])])
                                if not isinstance(t, Enumeration)
                                else TRUE,
                                common.sufficient_space_for_field_condition(
                                    Variable(f.affixed_name)
                                ),
                            )
                        ),
                        Postcondition(
                            And(
                                Call("Has_Buffer", [Variable("Ctx")]),
                                Call(
                                    "Valid",
                                    [Variable("Ctx"), Variable(f.affixed_name)],
                                ),
                                *(
                                    [
                                        Equal(
                                            Call(f"Get_{f.name}", [Variable("Ctx")]),
                                            Aggregate(TRUE, Variable("Val"))
                                            if isinstance(t, Enumeration) and t.always_valid
                                            else Variable("Val"),
                                        )
                                    ]
                                    if int(t.value_count) > 1
                                    else []
                                ),
                                *self.public_setter_postconditions(message, f),
                                *[
                                    Equal(
                                        Call(
                                            "Context_Cursor",
                                            [
                                                Variable("Ctx"),
                                                Variable(p.affixed_name),
                                            ],
                                        ),
                                        Old(
                                            Call(
                                                "Context_Cursor",
                                                [
                                                    Variable("Ctx"),
                                                    Variable(p.affixed_name),
                                                ],
                                            )
                                        ),
                                    )
                                    for p in message.predecessors(f)
                                ],
                            )
                        ),
                    ],
                )
                for f, t in scalar_fields.items()
            ],
            [
                SubprogramBody(
                    specification(f, t),
                    [
                        ObjectDeclaration(
                            ["Field_Value"],
                            "Field_Dependent_Value",
                            Aggregate(
                                Variable(f.affixed_name),
                                Call("To_Base", [Variable("Val")]),
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(["First", "Last"], const.TYPES_BIT_INDEX),
                    ],
                    [
                        CallStatement(
                            "Reset_Dependent_Fields",
                            [Variable("Ctx"), Variable(f.affixed_name)],
                        ),
                        CallStatement(
                            "Set_Field_Value",
                            [
                                Variable("Ctx"),
                                Variable("Field_Value"),
                                Variable("First"),
                                Variable("Last"),
                            ],
                        ),
                        self.__set_message_last(message, f),
                        Assignment(
                            Indexed(Variable("Ctx.Cursors"), Variable(f.affixed_name)),
                            NamedAggregate(
                                ("State", Variable("S_Valid")),
                                ("First", Variable("First")),
                                ("Last", Variable("Last")),
                                ("Value", Variable("Field_Value")),
                                (
                                    "Predecessor",
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "Predecessor",
                                    ),
                                ),
                            ),
                        ),
                        Assignment(
                            Indexed(
                                Variable("Ctx.Cursors"),
                                Call(
                                    "Successor",
                                    [Variable("Ctx"), Variable(f.affixed_name)],
                                ),
                            ),
                            NamedAggregate(
                                ("State", Variable("S_Invalid")),
                                ("Predecessor", Variable(f.affixed_name)),
                            ),
                        ),
                    ],
                )
                for f, t in scalar_fields.items()
            ],
        )

    def create_composite_setter_empty_procedures(self, message: Message) -> UnitPart:
        def specification(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Set_{field.name}_Empty", [InOutParameter(["Ctx"], "Context")]
            )

        return UnitPart(
            [
                *[
                    SubprogramDeclaration(
                        specification(f),
                        [
                            Precondition(
                                AndThen(
                                    *self.setter_preconditions(f),
                                    *self.composite_setter_field_condition_precondition(message, f),
                                    *self.composite_setter_preconditions(f),
                                    Equal(
                                        Call(
                                            "Field_Size",
                                            [
                                                Variable("Ctx"),
                                                Variable(f.affixed_name),
                                            ],
                                        ),
                                        Number(0),
                                    ),
                                )
                            ),
                            Postcondition(
                                And(
                                    *self.composite_setter_postconditions(f),
                                    *self.public_setter_postconditions(message, f),
                                )
                            ),
                        ],
                    )
                    for f, t in message.types.items()
                    if message.is_possibly_empty(f)
                ],
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        ObjectDeclaration(
                            ["First"],
                            const.TYPES_BIT_INDEX,
                            Call(
                                "Field_First",
                                [Variable("Ctx"), Variable(f.affixed_name)],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(
                            ["Last"],
                            const.TYPES_BIT_INDEX,
                            Call(
                                "Field_Last",
                                [Variable("Ctx"), Variable(f.affixed_name)],
                            ),
                            constant=True,
                        ),
                    ],
                    self.__set_sequence_field(message, f),
                )
                for f, t in message.types.items()
                if message.is_possibly_empty(f)
            ],
        )

    def create_sequence_setter_procedures(
        self, message: Message, sequence_fields: Mapping[Field, Type]
    ) -> UnitPart:
        def specification(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Set_{field.name}",
                [
                    InOutParameter(["Ctx"], "Context"),
                    Parameter(["Seq_Ctx"], f"{common.sequence_name(message, field)}.Context"),
                ],
            )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f),
                    [
                        Precondition(
                            AndThen(
                                *self.setter_preconditions(f),
                                *self.composite_setter_field_condition_precondition(message, f),
                                *self.composite_setter_preconditions(f),
                                Equal(
                                    Call(
                                        "Field_Size",
                                        [Variable("Ctx"), Variable(f.affixed_name)],
                                    ),
                                    Call(
                                        f"{common.sequence_name(message, f)}.Size",
                                        [Variable("Seq_Ctx")],
                                    ),
                                ),
                                Call(
                                    f"{common.sequence_name(message, f)}.Has_Buffer",
                                    [Variable("Seq_Ctx")],
                                ),
                                Call(
                                    f"{common.sequence_name(message, f)}.Valid",
                                    [Variable("Seq_Ctx")],
                                ),
                            )
                        ),
                        Postcondition(
                            And(
                                *self.composite_setter_postconditions(f),
                                *self.public_setter_postconditions(message, f),
                                If(
                                    [
                                        (
                                            Greater(
                                                Call(
                                                    "Field_Size",
                                                    [Variable("Ctx"), Variable(f.affixed_name)],
                                                ),
                                                Number(0),
                                            ),
                                            Call(
                                                "Present",
                                                [Variable("Ctx"), Variable(f.affixed_name)],
                                            ),
                                        )
                                    ]
                                ),
                            )
                        ),
                    ],
                )
                for f, t in sequence_fields.items()
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        *common.field_bit_location_declarations(Variable(f.affixed_name)),
                        *self.__field_byte_location_declarations(),
                    ],
                    [
                        *self.__set_sequence_field(message, f),
                        CallStatement(
                            f"{common.sequence_name(message, f)}.Copy",
                            [
                                Variable("Seq_Ctx"),
                                Indexed(
                                    Variable("Ctx.Buffer.all"),
                                    ValueRange(Variable("Buffer_First"), Variable("Buffer_Last")),
                                ),
                            ],
                        ),
                    ],
                )
                for f in sequence_fields
            ],
        )

    def create_opaque_setter_procedures(self, message: Message) -> UnitPart:
        def specification(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Set_{field.name}",
                [InOutParameter(["Ctx"], "Context"), Parameter(["Data"], const.TYPES_BYTES)],
            )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f),
                    [
                        Precondition(
                            AndThen(
                                *self.setter_preconditions(f),
                                *self.composite_setter_preconditions(f),
                                Equal(
                                    Length("Data"),
                                    Call(
                                        const.TYPES_TO_LENGTH,
                                        [
                                            Call(
                                                "Field_Size",
                                                [
                                                    Variable("Ctx"),
                                                    Variable(f.affixed_name),
                                                ],
                                            )
                                        ],
                                    ),
                                ),
                                *self.composite_setter_field_condition_precondition(message, f),
                            )
                        ),
                        Postcondition(
                            And(
                                *self.composite_setter_postconditions(f),
                                *self.public_setter_postconditions(message, f),
                            )
                        ),
                    ],
                )
                for f, t in message.types.items()
                if isinstance(t, Opaque)
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        *common.field_bit_location_declarations(Variable(f.affixed_name)),
                        *self.__field_byte_location_declarations(),
                    ],
                    [
                        CallStatement(f"Initialize_{f.name}_Private", [Variable("Ctx")]),
                        Assignment(
                            Slice(
                                Selected(Variable("Ctx.Buffer"), "all"),
                                Variable("Buffer_First"),
                                Variable("Buffer_Last"),
                            ),
                            Variable("Data"),
                        ),
                    ],
                )
                for f, t in message.types.items()
                if isinstance(t, Opaque)
            ],
        )

    def create_generic_opaque_setter_procedures(self, message: Message) -> UnitPart:
        def specification(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Generic_Set_{field.name}", [InOutParameter(["Ctx"], "Context")]
            )

        def formal_parameters(field: Field) -> Sequence[FormalSubprogramDeclaration]:
            return [
                FormalSubprogramDeclaration(
                    ProcedureSpecification(
                        f"Process_{field.name}",
                        [OutParameter([field.name], const.TYPES_BYTES)],
                    )
                ),
                FormalSubprogramDeclaration(
                    FunctionSpecification(
                        "Valid_Length",
                        "Boolean",
                        [Parameter(["Length"], const.TYPES_LENGTH)],
                    )
                ),
            ]

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f),
                    [
                        Precondition(
                            AndThen(
                                *self.setter_preconditions(f),
                                *self.composite_setter_preconditions(f),
                                Call(
                                    "Valid_Length",
                                    [
                                        Call(
                                            const.TYPES_LENGTH,
                                            [
                                                Div(
                                                    Call(
                                                        "Field_Size",
                                                        [
                                                            Variable("Ctx"),
                                                            Variable(f.affixed_name),
                                                        ],
                                                    ),
                                                    Size(const.TYPES_BYTE),
                                                ),
                                            ],
                                        ),
                                    ],
                                ),
                            )
                        ),
                        Postcondition(
                            And(
                                *self.composite_setter_postconditions(f),
                                *self.public_setter_postconditions(message, f),
                            )
                        ),
                    ],
                    formal_parameters(f),
                )
                for f, t in message.types.items()
                if isinstance(t, Opaque)
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        *common.field_bit_location_declarations(Variable(f.affixed_name)),
                        *self.__field_byte_location_declarations(),
                    ],
                    [
                        CallStatement(f"Initialize_{f.name}_Private", [Variable("Ctx")]),
                        CallStatement(
                            f"Process_{f.name}",
                            [
                                Slice(
                                    Selected(Variable("Ctx.Buffer"), "all"),
                                    Variable("Buffer_First"),
                                    Variable("Buffer_Last"),
                                ),
                            ],
                        ),
                    ],
                )
                for f, t in message.types.items()
                if isinstance(t, Opaque)
            ],
        )

    def create_composite_initialize_procedures(self, message: Message) -> UnitPart:
        def specification_private(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Initialize_{field.name}_Private", [InOutParameter(["Ctx"], "Context")]
            )

        def specification_public(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Initialize_{field.name}", [InOutParameter(["Ctx"], "Context")]
            )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification_public(f),
                    [
                        Precondition(
                            AndThen(
                                *self.setter_preconditions(f),
                                *self.composite_setter_preconditions(f),
                            )
                        ),
                        Postcondition(
                            And(
                                *self.composite_setter_postconditions(f),
                                *self.public_setter_postconditions(message, f),
                            )
                        ),
                    ],
                )
                for f, t in message.types.items()
                if isinstance(t, Opaque)
            ],
            [
                s
                for f, t in message.types.items()
                if isinstance(t, Opaque)
                for s in [
                    SubprogramBody(
                        specification_private(f),
                        common.field_bit_location_declarations(Variable(f.affixed_name)),
                        common.initialize_field_statements(message, f),
                        [
                            Precondition(
                                AndThen(
                                    *self.setter_preconditions(f),
                                    *self.composite_setter_preconditions(f),
                                )
                            ),
                            Postcondition(
                                And(
                                    *self.composite_setter_postconditions(f),
                                    *self.private_setter_postconditions(message, f),
                                )
                            ),
                        ],
                    ),
                    SubprogramBody(
                        specification_public(f),
                        [],
                        [
                            CallStatement(specification_private(f).identifier, [Variable("Ctx")]),
                        ],
                    ),
                ]
            ],
        )

    @staticmethod
    def setter_preconditions(field: Field) -> Sequence[Expr]:
        return [
            Not(Constrained("Ctx")),
            Call("Has_Buffer", [Variable("Ctx")]),
            Call("Valid_Next", [Variable("Ctx"), Variable(field.affixed_name)]),
        ]

    def private_setter_postconditions(self, message: Message, field: Field) -> Sequence[Expr]:
        return [
            Equal(
                Variable("Ctx.Message_Last"),
                Call(
                    "Field_Last",
                    [Variable("Ctx"), Variable(field.affixed_name)],
                ),
            ),
            *self.setter_postconditions(message, field),
        ]

    def public_setter_postconditions(self, message: Message, field: Field) -> Sequence[Expr]:
        return [
            *(
                [
                    If(
                        [
                            (
                                Call("Structural_Valid_Message", [Variable("Ctx")]),
                                Equal(
                                    Call("Message_Last", [Variable("Ctx")]),
                                    Call(
                                        "Field_Last",
                                        [Variable("Ctx"), Variable(field.affixed_name)],
                                    ),
                                ),
                            )
                        ]
                    )
                ]
                if field in message.direct_predecessors(FINAL)
                else []
            ),
            *self.setter_postconditions(message, field),
        ]

    def setter_postconditions(self, message: Message, field: Field) -> Sequence[Expr]:
        return [
            *[
                Call("Invalid", [Variable("Ctx"), Variable(p.affixed_name)])
                for p in message.successors(field)
                if p != FINAL
            ],
            *common.valid_path_to_next_field_condition(message, field, self.prefix),
            *const.CONTEXT_INVARIANT,
            *[
                Equal(e, Old(e))
                for e in [
                    Call(
                        "Predecessor",
                        [Variable("Ctx"), Variable(field.affixed_name)],
                    ),
                    Call(
                        "Valid_Next",
                        [Variable("Ctx"), Variable(field.affixed_name)],
                    ),
                ]
                + [
                    Call(f"Get_{p.name}", [Variable("Ctx")])
                    for p in message.definite_predecessors(field)
                    for t in [message.types[p]]
                    if isinstance(t, Scalar) and int(t.value_count) > 1
                ]
            ],
        ]

    @staticmethod
    def composite_setter_postconditions(field: Field) -> Sequence[Expr]:
        return [
            Call("Has_Buffer", [Variable("Ctx")]),
            Call("Structural_Valid", [Variable("Ctx"), Variable(field.affixed_name)]),
        ]

    @staticmethod
    def composite_setter_field_condition_precondition(
        message: Message, field: Field
    ) -> Sequence[Expr]:
        return [
            common.field_condition_call(message, field, Variable("Data")),
        ]

    @staticmethod
    def composite_setter_preconditions(field: Field) -> Sequence[Expr]:
        return [
            common.sufficient_space_for_field_condition(Variable(field.affixed_name)),
            Equal(
                Mod(
                    Call(
                        "Field_First",
                        [Variable("Ctx"), Variable(field.affixed_name)],
                    ),
                    Size(const.TYPES_BYTE),
                ),
                Number(1),
            ),
            Equal(
                Mod(
                    Call(
                        "Field_Last",
                        [Variable("Ctx"), Variable(field.affixed_name)],
                    ),
                    Size(const.TYPES_BYTE),
                ),
                Number(0),
            ),
            Equal(
                Mod(
                    Call(
                        "Field_Size",
                        [Variable("Ctx"), Variable(field.affixed_name)],
                    ),
                    Size(const.TYPES_BYTE),
                ),
                Number(0),
            ),
        ]

    @staticmethod
    def __set_message_last(message: Message, field: Field) -> Statement:
        return (
            Assignment(Variable("Ctx.Message_Last"), Variable("Last"))
            if field in message.direct_predecessors(FINAL)
            else Assignment(
                Variable("Ctx.Message_Last"),
                Mul(
                    Div(
                        Add(
                            Variable("Last"),
                            Number(7),
                        ),
                        Number(8),
                    ),
                    Number(8),
                ),
            )
        )

    def __set_sequence_field(self, message: Message, field: Field) -> Sequence[Statement]:
        return [
            CallStatement(
                "Reset_Dependent_Fields",
                [Variable("Ctx"), Variable(field.affixed_name)],
            ),
            self.__set_message_last(message, field),
            Assignment(
                Indexed(Variable("Ctx.Cursors"), Variable(field.affixed_name)),
                NamedAggregate(
                    ("State", Variable("S_Valid")),
                    ("First", Variable("First")),
                    ("Last", Variable("Last")),
                    (
                        "Value",
                        NamedAggregate(("Fld", Variable(field.affixed_name))),
                    ),
                    (
                        "Predecessor",
                        Selected(
                            Indexed(
                                Variable("Ctx.Cursors"),
                                Variable(field.affixed_name),
                            ),
                            "Predecessor",
                        ),
                    ),
                ),
            ),
            Assignment(
                Indexed(
                    Variable("Ctx.Cursors"),
                    Call(
                        "Successor",
                        [Variable("Ctx"), Variable(field.affixed_name)],
                    ),
                ),
                NamedAggregate(
                    ("State", Variable("S_Invalid")),
                    ("Predecessor", Variable(field.affixed_name)),
                ),
            ),
        ]

    @staticmethod
    def __field_byte_location_declarations() -> Sequence[Declaration]:
        return [
            ExpressionFunctionDeclaration(
                FunctionSpecification("Buffer_First", const.TYPES_INDEX),
                Call(const.TYPES_TO_INDEX, [Variable("First")]),
            ),
            ExpressionFunctionDeclaration(
                FunctionSpecification("Buffer_Last", const.TYPES_INDEX),
                Call(const.TYPES_TO_INDEX, [Variable("Last")]),
            ),
        ]



from typing import Callable, List, Mapping, Optional, Sequence, Tuple

from rflx import ada, expression as expr, identifier as rid, model
from rflx.const import BUILTINS_PACKAGE

from . import const


def substitution(
    message: model.Message,
    prefix: str,
    embedded: bool = False,
    public: bool = False,
    target_type: Optional[ada.ID] = const.TYPES_U64,
) -> Callable[[expr.Expr], expr.Expr]:
    facts = substitution_facts(message, prefix, embedded, public, target_type)

    def func(expression: expr.Expr) -> expr.Expr:
        def byte_aggregate(aggregate: expr.Aggregate) -> expr.Aggregate:
            return expr.Aggregate(*[expr.Val(const.TYPES_BYTE, e) for e in aggregate.elements])

        if isinstance(expression, expr.Name) and expression in facts:
            return facts[expression]

        if isinstance(expression, expr.String):
            return expr.Aggregate(*expression.elements)

        if isinstance(expression, (expr.Equal, expr.NotEqual)):
            field = None
            aggregate = None
            if isinstance(expression.left, expr.Variable) and isinstance(
                expression.right, expr.Aggregate
            ):
                field = model.Field(expression.left.name)
                aggregate = byte_aggregate(expression.right)
            elif isinstance(expression.left, expr.Aggregate) and isinstance(
                expression.right, expr.Variable
            ):
                field = model.Field(expression.right.name)
                aggregate = byte_aggregate(expression.left)
            if field and len(field.identifier.parts) == 1 and aggregate:
                assert field in message.fields
                if embedded:
                    return expression.__class__(
                        expr.Indexed(
                            expr.Variable(expr.ID("Buffer") * "all"),
                            expr.ValueRange(
                                expr.Call(
                                    const.TYPES_TO_INDEX,
                                    [
                                        expr.Selected(
                                            expr.Indexed(
                                                expr.Variable("Cursors"),
                                                expr.Variable(field.affixed_name),
                                            ),
                                            "First",
                                        )
                                    ],
                                ),
                                expr.Call(
                                    const.TYPES_TO_INDEX,
                                    [
                                        expr.Selected(
                                            expr.Indexed(
                                                expr.Variable("Cursors"),
                                                expr.Variable(field.affixed_name),
                                            ),
                                            "Last",
                                        )
                                    ],
                                ),
                            ),
                        ),
                        aggregate,
                    )
                equal_call = expr.Call(
                    "Equal",
                    [expr.Variable("Ctx"), expr.Variable(field.affixed_name), aggregate],
                )
                return equal_call if isinstance(expression, expr.Equal) else expr.Not(equal_call)

        def field_value(field: model.Field) -> expr.Expr:
            if public:
                return expr.Call(f"Get_{field.name}", [expr.Variable("Ctx")])
            return expr.Selected(
                expr.Indexed(
                    expr.Variable(expr.ID("Ctx") * "Cursors" if not embedded else "Cursors"),
                    expr.Variable(field.affixed_name),
                ),
                expr.ID("Value") * f"{field.name}_Value",
            )

        if isinstance(expression, expr.Relation):
            if (
                isinstance(expression.left, expr.Variable)
                and model.Field(expression.left.name) in message.fields
                and isinstance(expression.right, expr.Number)
            ):
                return expression.__class__(
                    field_value(model.Field(expression.left.name)), expression.right
                )
            if (
                isinstance(expression.right, expr.Variable)
                and model.Field(expression.right.name) in message.fields
                and isinstance(expression.left, expr.Number)
            ):
                return expression.__class__(
                    expression.left, field_value(model.Field(expression.right.name))
                )

        return expression

    return func


def substitution_facts(
    message: model.Message,
    prefix: str,
    embedded: bool = False,
    public: bool = False,
    target_type: Optional[ada.ID] = const.TYPES_U64,
) -> Mapping[expr.Name, expr.Expr]:
    def prefixed(name: str) -> expr.Expr:
        return expr.Variable(expr.ID("Ctx") * name) if not embedded else expr.Variable(name)

    first = prefixed("First")
    last = prefixed("Last")
    cursors = prefixed("Cursors")

    def field_first(field: model.Field) -> expr.Expr:
        if public:
            return expr.Call(
                "Field_First", [expr.Variable("Ctx"), expr.Variable(field.affixed_name)]
            )
        return expr.Selected(expr.Indexed(cursors, expr.Variable(field.affixed_name)), "First")

    def field_last(field: model.Field) -> expr.Expr:
        if public:
            return expr.Call(
                "Field_Last", [expr.Variable("Ctx"), expr.Variable(field.affixed_name)]
            )
        return expr.Selected(expr.Indexed(cursors, expr.Variable(field.affixed_name)), "Last")

    def field_size(field: model.Field) -> expr.Expr:
        if public:
            return expr.Call(
                "Field_Size", [expr.Variable("Ctx"), expr.Variable(field.affixed_name)]
            )
        return expr.Add(
            expr.Sub(
                expr.Selected(expr.Indexed(cursors, expr.Variable(field.affixed_name)), "Last"),
                expr.Selected(expr.Indexed(cursors, expr.Variable(field.affixed_name)), "First"),
            ),
            expr.Number(1),
        )

    def field_value(field: model.Field, field_type: model.Type) -> expr.Expr:
        if isinstance(field_type, model.Enumeration):
            if public:
                return expr.Call(
                    "To_Base", [expr.Call(f"Get_{field.name}", [expr.Variable("Ctx")])]
                )
            return expr.Selected(
                expr.Indexed(cursors, expr.Variable(field.affixed_name)),
                expr.ID("Value") * f"{field.name}_Value",
            )
        if isinstance(field_type, model.Scalar):
            if public:
                return expr.Call(f"Get_{field.name}", [expr.Variable("Ctx")])
            return expr.Selected(
                expr.Indexed(cursors, expr.Variable(field.affixed_name)),
                expr.ID("Value") * f"{field.name}_Value",
            )
        if isinstance(field_type, model.Composite):
            return expr.Variable(field.name)

        assert False, f'unexpected type "{type(field_type).__name__}"'

    def type_conversion(expression: expr.Expr) -> expr.Expr:
        return expr.Call(target_type, [expression]) if target_type else expression

    return {
        **{expr.First("Message"): type_conversion(first)},
        **{expr.Last("Message"): type_conversion(last)},
        **{expr.Size("Message"): type_conversion(expr.Add(last, -first, expr.Number(1)))},
        **{expr.First(f.name): type_conversion(field_first(f)) for f in message.fields},
        **{expr.Last(f.name): type_conversion(field_last(f)) for f in message.fields},
        **{expr.Size(f.name): type_conversion(field_size(f)) for f in message.fields},
        **{
            expr.Variable(f.name): type_conversion(field_value(f, t))
            for f, t in message.types.items()
        },
        **{
            expr.Variable(l): type_conversion(expr.Call("To_Base", [expr.Variable(l)]))
            for t in message.types.values()
            if isinstance(t, model.Enumeration)
            for l in t.literals.keys()
        },
        **{
            expr.Variable(t.package * l): type_conversion(
                expr.Call("To_Base", [expr.Variable(prefix * t.package * l)])
            )
            for t in message.types.values()
            if isinstance(t, model.Enumeration)
            for l in t.literals.keys()
        },
        # ISSUE: Componolit/RecordFlux#276
        **{expr.ValidChecksum(f): expr.TRUE for f in message.checksums},
    }


def message_structure_invariant(
    message: model.Message, prefix: str, link: model.Link = None, embedded: bool = False
) -> ada.Expr:
    def prefixed(name: str) -> expr.Expr:
        return expr.Selected(expr.Variable("Ctx"), name) if not embedded else expr.Variable(name)

    if not link:
        return message_structure_invariant(
            message, prefix, message.outgoing(model.INITIAL)[0], embedded
        )

    source = link.source
    target = link.target

    if target is model.FINAL:
        return ada.TRUE

    field_type = message.types[target]
    condition = link.condition.substituted(substitution(message, prefix, embedded)).simplified()
    size = (
        expr.Size(prefix * full_base_type_name(field_type))
        if isinstance(field_type, model.Scalar)
        else link.size.substituted(
            substitution(message, prefix, embedded, target_type=const.TYPES_BIT_LENGTH)
        ).simplified()
    )
    first = (
        prefixed("First")
        if source == model.INITIAL
        else link.first.substituted(
            substitution(message, prefix, embedded, target_type=const.TYPES_BIT_INDEX)
        )
        .substituted(
            mapping={
                expr.UNDEFINED: expr.Add(
                    expr.Selected(
                        expr.Indexed(prefixed("Cursors"), expr.Variable(source.affixed_name)),
                        "Last",
                    ),
                    expr.Number(1),
                )
            }
        )
        .simplified()
    )
    invariant = [
        message_structure_invariant(message, prefix, l, embedded) for l in message.outgoing(target)
    ]

    return ada.If(
        [
            (
                ada.AndThen(
                    ada.Call(
                        "Structural_Valid",
                        [
                            ada.Indexed(
                                prefixed("Cursors").ada_expr(), ada.Variable(target.affixed_name)
                            )
                        ],
                    ),
                    *([condition.ada_expr()] if condition != expr.TRUE else []),
                ),
                ada.AndThen(
                    ada.Equal(
                        ada.Add(
                            ada.Sub(
                                ada.Selected(
                                    ada.Indexed(
                                        prefixed("Cursors").ada_expr(),
                                        ada.Variable(target.affixed_name),
                                    ),
                                    "Last",
                                ),
                                ada.Selected(
                                    ada.Indexed(
                                        prefixed("Cursors").ada_expr(),
                                        ada.Variable(target.affixed_name),
                                    ),
                                    "First",
                                ),
                            ),
                            ada.Number(1),
                        ),
                        size.ada_expr(),
                    ),
                    ada.Equal(
                        ada.Selected(
                            ada.Indexed(
                                prefixed("Cursors").ada_expr(),
                                ada.Variable(target.affixed_name),
                            ),
                            "Predecessor",
                        ),
                        ada.Variable(source.affixed_name),
                    ),
                    ada.Equal(
                        ada.Selected(
                            ada.Indexed(
                                prefixed("Cursors").ada_expr(),
                                ada.Variable(target.affixed_name),
                            ),
                            "First",
                        ),
                        first.ada_expr(),
                    ),
                    *[i for i in invariant if i != ada.TRUE],
                ),
            )
        ]
    )


def context_predicate(
    message: model.Message, composite_fields: Sequence[model.Field], prefix: str
) -> ada.Expr:
    def valid_predecessors_invariant() -> ada.Expr:
        return ada.AndThen(
            *[
                ada.If(
                    [
                        (
                            ada.Call(
                                "Structural_Valid",
                                [
                                    ada.Indexed(
                                        ada.Variable("Cursors"),
                                        ada.Variable(f.affixed_name),
                                    )
                                ],
                            ),
                            ada.Or(
                                *[
                                    expr.AndThen(
                                        expr.Call(
                                            "Structural_Valid"
                                            if l.source in composite_fields
                                            else "Valid",
                                            [
                                                expr.Indexed(
                                                    expr.Variable("Cursors"),
                                                    expr.Variable(l.source.affixed_name),
                                                )
                                            ],
                                        ),
                                        expr.Equal(
                                            expr.Selected(
                                                expr.Indexed(
                                                    expr.Variable("Cursors"),
                                                    expr.Variable(f.affixed_name),
                                                ),
                                                "Predecessor",
                                            ),
                                            expr.Variable(l.source.affixed_name),
                                        ),
                                        l.condition.substituted(
                                            substitution(message, embedded=True, prefix=prefix)
                                        ),
                                    )
                                    .simplified()
                                    .ada_expr()
                                    for l in message.incoming(f)
                                ]
                            ),
                        )
                    ]
                )
                for f in message.fields
                if f not in message.direct_successors(model.INITIAL)
            ]
        )

    def invalid_successors_invariant() -> ada.Expr:
        return ada.AndThen(
            *[
                ada.If(
                    [
                        (
                            ada.AndThen(
                                *[
                                    ada.Call(
                                        "Invalid",
                                        [
                                            ada.Indexed(
                                                ada.Variable("Cursors"),
                                                ada.Variable(p.affixed_name),
                                            )
                                        ],
                                    )
                                    for p in message.direct_predecessors(f)
                                ]
                            ),
                            ada.Call(
                                "Invalid",
                                [
                                    ada.Indexed(
                                        ada.Variable("Cursors"),
                                        ada.Variable(f.affixed_name),
                                    )
                                ],
                            ),
                        )
                    ]
                )
                for f in message.fields
                if f not in message.direct_successors(model.INITIAL)
            ]
        )

    return ada.AndThen(
        ada.If(
            [
                (
                    ada.NotEqual(ada.Variable("Buffer"), ada.Variable("null")),
                    ada.And(
                        ada.Equal(ada.First("Buffer"), ada.Variable("Buffer_First")),
                        ada.Equal(ada.Last("Buffer"), ada.Variable("Buffer_Last")),
                    ),
                )
            ]
        ),
        public_context_predicate(),
        ada.LessEqual(ada.Sub(ada.Variable("First"), ada.Number(1)), ada.Variable("Message_Last")),
        ada.LessEqual(ada.Variable("Message_Last"), ada.Variable("Last")),
        ada.Equal(ada.Mod(ada.Variable("First"), ada.Size(const.TYPES_BYTE)), ada.Number(1)),
        ada.Equal(ada.Mod(ada.Variable("Last"), ada.Size(const.TYPES_BYTE)), ada.Number(0)),
        ada.Equal(ada.Mod(ada.Variable("Message_Last"), ada.Size(const.TYPES_BYTE)), ada.Number(0)),
        ada.ForAllIn(
            "F",
            ada.ValueRange(ada.First("Field"), ada.Last("Field")),
            ada.If(
                [
                    (
                        ada.Call(
                            "Structural_Valid",
                            [ada.Indexed(ada.Variable("Cursors"), ada.Variable("F"))],
                        ),
                        ada.And(
                            ada.GreaterEqual(
                                ada.Selected(
                                    ada.Indexed(ada.Variable("Cursors"), ada.Variable("F")), "First"
                                ),
                                ada.Variable("First"),
                            ),
                            ada.LessEqual(
                                ada.Selected(
                                    ada.Indexed(ada.Variable("Cursors"), ada.Variable("F")), "Last"
                                ),
                                ada.Variable("Message_Last"),
                            ),
                            ada.LessEqual(
                                ada.Selected(
                                    ada.Indexed(ada.Variable("Cursors"), ada.Variable("F")), "First"
                                ),
                                ada.Add(
                                    ada.Selected(
                                        ada.Indexed(ada.Variable("Cursors"), ada.Variable("F")),
                                        "Last",
                                    ),
                                    ada.Number(1),
                                ),
                            ),
                            ada.Equal(
                                ada.Selected(
                                    ada.Selected(
                                        ada.Indexed(ada.Variable("Cursors"), ada.Variable("F")),
                                        "Value",
                                    ),
                                    "Fld",
                                ),
                                ada.Variable("F"),
                            ),
                        ),
                    )
                ]
            ),
        ),
        valid_predecessors_invariant(),
        invalid_successors_invariant(),
        message_structure_invariant(message, prefix, embedded=True),
    )


def public_context_predicate() -> ada.Expr:
    return ada.And(
        ada.GreaterEqual(
            ada.Call(const.TYPES_TO_INDEX, [ada.Variable("First")]), ada.Variable("Buffer_First")
        ),
        ada.LessEqual(
            ada.Call(const.TYPES_TO_INDEX, [ada.Variable("Last")]), ada.Variable("Buffer_Last")
        ),
        ada.Less(ada.Variable("Buffer_Last"), ada.Last(const.TYPES_INDEX)),
        ada.LessEqual(ada.Variable("First"), ada.Add(ada.Variable("Last"), ada.Number(1))),
        ada.Less(ada.Variable("Last"), ada.Last(const.TYPES_BIT_INDEX)),
        ada.Equal(ada.Mod(ada.Variable("First"), ada.Size(const.TYPES_BYTE)), ada.Number(1)),
        ada.Equal(ada.Mod(ada.Variable("Last"), ada.Size(const.TYPES_BYTE)), ada.Number(0)),
    )


def valid_path_to_next_field_condition(
    message: model.Message, field: model.Field, prefix: str
) -> Sequence[ada.Expr]:
    return [
        ada.If(
            [
                (
                    l.condition.substituted(substitution(message, public=True, prefix=prefix))
                    .simplified()
                    .ada_expr(),
                    ada.And(
                        ada.Equal(
                            ada.Call(
                                "Predecessor",
                                [ada.Variable("Ctx"), ada.Variable(l.target.affixed_name)],
                            ),
                            ada.Variable(field.affixed_name),
                        ),
                        ada.Call(
                            "Valid_Next",
                            [ada.Variable("Ctx"), ada.Variable(l.target.affixed_name)],
                        )
                        if l.target != model.FINAL
                        else ada.TRUE,
                    ),
                )
            ]
        )
        for l in message.outgoing(field)
        if l.target != model.FINAL
    ]


def sufficient_space_for_field_condition(field_name: ada.Name) -> ada.Expr:
    return ada.GreaterEqual(
        ada.Call("Available_Space", [ada.Variable("Ctx"), field_name]),
        ada.Call("Field_Size", [ada.Variable("Ctx"), field_name]),
    )


def initialize_field_statements(
    message: model.Message, field: model.Field
) -> Sequence[ada.Statement]:
    comparison_to_aggregate = is_compared_to_aggregate(field, message)

    if comparison_to_aggregate:
        size = message.field_size(field)
        assert isinstance(size, expr.Number)
        aggregate_size = int(size) // 8

    return [
        ada.CallStatement(
            "Reset_Dependent_Fields",
            [ada.Variable("Ctx"), ada.Variable(field.affixed_name)],
        ),
        ada.Assignment("Ctx.Message_Last", ada.Variable("Last")),
        ada.Assignment(
            ada.Indexed(ada.Variable("Ctx.Cursors"), ada.Variable(field.affixed_name)),
            ada.NamedAggregate(
                ("State", ada.Variable("S_Structural_Valid")),
                ("First", ada.Variable("First")),
                ("Last", ada.Variable("Last")),
                (
                    "Value",
                    ada.NamedAggregate(
                        ("Fld", ada.Variable(field.affixed_name)),
                        *(
                            [
                                (
                                    f"{field.identifier}_Value",
                                    ada.Aggregate(*([ada.Number(0)] * aggregate_size)),
                                )
                            ]
                            if comparison_to_aggregate
                            else []
                        ),
                    ),
                ),
                (
                    "Predecessor",
                    ada.Selected(
                        ada.Indexed(
                            ada.Variable("Ctx.Cursors"),
                            ada.Variable(field.affixed_name),
                        ),
                        "Predecessor",
                    ),
                ),
            ),
        ),
        ada.Assignment(
            ada.Indexed(
                ada.Variable("Ctx.Cursors"),
                ada.Call(
                    "Successor",
                    [ada.Variable("Ctx"), ada.Variable(field.affixed_name)],
                ),
            ),
            ada.NamedAggregate(
                ("State", ada.Variable("S_Invalid")),
                ("Predecessor", ada.Variable(field.affixed_name)),
            ),
        ),
    ]


def field_bit_location_declarations(field_name: ada.Name) -> Sequence[ada.Declaration]:
    return [
        ada.ObjectDeclaration(
            ["First"],
            const.TYPES_BIT_INDEX,
            ada.Call("Field_First", [ada.Variable("Ctx"), field_name]),
            constant=True,
        ),
        ada.ObjectDeclaration(
            ["Last"],
            const.TYPES_BIT_INDEX,
            ada.Call("Field_Last", [ada.Variable("Ctx"), field_name]),
            constant=True,
        ),
    ]


def field_byte_bounds_declarations() -> Sequence[ada.Declaration]:
    return [
        ada.ExpressionFunctionDeclaration(
            ada.FunctionSpecification("Buffer_First", const.TYPES_INDEX),
            ada.Call(const.TYPES_TO_INDEX, [ada.Variable("First")]),
        ),
        ada.ExpressionFunctionDeclaration(
            ada.FunctionSpecification("Buffer_Last", const.TYPES_INDEX),
            ada.Call(const.TYPES_TO_INDEX, [ada.Variable("Last")]),
        ),
    ]


def field_byte_location_declarations() -> Sequence[ada.Declaration]:
    return [
        *field_byte_bounds_declarations(),
        ada.ExpressionFunctionDeclaration(
            ada.FunctionSpecification("Offset", const.TYPES_OFFSET),
            ada.Call(
                const.TYPES_OFFSET,
                [
                    ada.Mod(
                        ada.Sub(ada.Number(8), ada.Mod(ada.Variable("Last"), ada.Number(8))),
                        ada.Number(8),
                    )
                ],
            ),
        ),
    ]


def field_condition_call(message: model.Message, field: model.Field, value: ada.Expr) -> ada.Expr:
    return ada.Call(
        "Field_Condition",
        [
            ada.Variable("Ctx"),
            ada.Call(
                f"Construct_{field.identifier}_Value", [value]
            )  # ISSUE: Componolit/Workarounds#35
            if is_compared_to_aggregate(field, message)
            else ada.NamedAggregate(
                ("Fld", ada.Variable(field.affixed_name)),
                *(
                    [(f"{field.identifier}_Value", value)]
                    if isinstance(message.types[field], model.Scalar)
                    else []
                ),
            ),
            *(
                [
                    ada.Call(
                        "Field_Size",
                        [
                            ada.Variable("Ctx"),
                            ada.Variable(field.affixed_name),
                        ],
                    )
                ]
                if size_dependent_condition(message)
                else []
            ),
        ],
    )


def prefixed_type_identifier(type_identifier: ada.ID, prefix: str) -> ada.ID:
    if model.is_builtin_type(type_identifier):
        return type_identifier.name

    return prefix * type_identifier


def base_type_name(scalar_type: model.Scalar) -> ada.ID:
    if isinstance(scalar_type, model.ModularInteger):
        return ada.ID(scalar_type.name)

    return ada.ID(scalar_type.name + "_Base")


def full_base_type_name(scalar_type: model.Scalar) -> ada.ID:
    if scalar_type.package == BUILTINS_PACKAGE:
        return const.BUILTIN_TYPES_PACKAGE * scalar_type.name + "_Base"

    if isinstance(scalar_type, model.ModularInteger):
        return ada.ID(scalar_type.identifier)

    return ada.ID(scalar_type.identifier + "_Base")


def enum_name(enum_type: model.Enumeration) -> ada.ID:
    return ada.ID(enum_type.name + "_Enum")


def full_enum_name(enum_type: model.Enumeration) -> ada.ID:
    return ada.ID(enum_type.identifier + "_Enum")


def sequence_name(message: model.Message, field: model.Field) -> ada.ID:
    return ada.ID(message.types[field].identifier)


def contains_function_name(
    refinement_package: rid.ID, pdu: rid.ID, sdu: rid.ID, field: rid.ID
) -> str:
    sdu_name = sdu.name if sdu.parent == refinement_package else sdu
    pdu_name = pdu.name if pdu.parent == refinement_package else pdu
    return f"{sdu_name.flat}_In_{pdu_name.flat}_{field}"


def size_dependent_condition(message: model.Message) -> bool:
    field_sizes = {expr.Size(f.name) for f in message.fields}
    return any(
        size in field_sizes
        for link in message.structure
        for size in link.condition.findall(lambda x: isinstance(x, expr.Size))
    )


def is_compared_to_aggregate(field: model.Field, message: model.Message) -> bool:
    return any(
        r
        for l in message.structure
        for r in l.condition.findall(lambda x: isinstance(x, (expr.Equal, expr.NotEqual)))
        if isinstance(r, (expr.Equal, expr.NotEqual))
        and any(r.findall(lambda x: isinstance(x, expr.Aggregate)))
        and (
            r.left == expr.Variable(field.identifier) or r.right == expr.Variable(field.identifier)
        )
    )


def create_sequence_instantiation(
    sequence_type: model.Sequence,
    prefix: str = "",
    flat: bool = False,
) -> Tuple[List[ada.ContextItem], ada.GenericPackageInstantiation]:
    element_type = sequence_type.element_type
    element_type_package = ada.ID(element_type.package.name)

    sequence_context: List[ada.ContextItem] = []
    sequence_package: ada.GenericPackageInstantiation
    if isinstance(element_type, model.Message):
        element_type_identifier = ada.ID(
            element_type.identifier.flat if flat else prefix * element_type.identifier
        )
        sequence_context = [
            ada.WithClause(prefix * const.MESSAGE_SEQUENCE_PACKAGE),
            *([] if flat else [ada.WithClause(element_type_identifier)]),
        ]
        sequence_package = ada.GenericPackageInstantiation(
            ada.ID(sequence_type.identifier.flat if flat else prefix * sequence_type.identifier),
            prefix * const.MESSAGE_SEQUENCE_PACKAGE,
            [
                element_type_identifier * "Context",
                element_type_identifier * "Initialize",
                element_type_identifier * "Take_Buffer",
                element_type_identifier * "Has_Buffer",
                element_type_identifier * "Message_Last",
                element_type_identifier * "Initialized",
                element_type_identifier * "Structural_Valid_Message",
            ],
        )
    elif isinstance(element_type, model.Scalar):
        element_type_identifier = ada.ID(prefix * element_type.identifier)
        sequence_context = [
            ada.WithClause(prefix * const.SCALAR_SEQUENCE_PACKAGE),
            ada.WithClause(prefix * element_type_package),
        ]
        sequence_package = ada.GenericPackageInstantiation(
            ada.ID(sequence_type.identifier.flat if flat else prefix * sequence_type.identifier),
            prefix * const.SCALAR_SEQUENCE_PACKAGE,
            [
                element_type_identifier,
                prefix
                * element_type_package
                * (
                    base_type_name(element_type)
                    if not isinstance(element_type, model.ModularInteger)
                    else element_type.name
                ),
                prefix * element_type_package * "Valid",
                prefix * element_type_package * "To_Actual",
                prefix * element_type_package * "To_Base",
            ],
        )
    else:
        assert False, 'unexpected element type "{type(element_type)}"'

    return (sequence_context, sequence_package)



import argparse
import json
import logging
import os
import traceback
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Sequence, Union

import librflxlang
from pkg_resources import get_distribution

from rflx import __version__
from rflx.error import RecordFluxError, Severity, Subsystem, fail
from rflx.generator import Generator
from rflx.graph import Graph
from rflx.model import Message, Model, Session
from rflx.specification import Parser

logging.basicConfig(level=logging.INFO, format="%(message)s")

DEFAULT_PREFIX = "RFLX"


def main(argv: List[str]) -> Union[int, str]:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="disable logging to standard output"
    )
    parser.add_argument("--version", action="store_true")

    subparsers = parser.add_subparsers(dest="subcommand")

    parser_check = subparsers.add_parser("check", help="check specification")
    parser_check.add_argument(
        "files", metavar="FILE", type=Path, nargs="+", help="specification file"
    )
    parser_check.set_defaults(func=check)

    parser_generate = subparsers.add_parser("generate", help="generate code")
    parser_generate.add_argument(
        "-p",
        "--prefix",
        type=str,
        default="RFLX",
        help=("add prefix to generated packages " f"(default: {DEFAULT_PREFIX})"),
    )
    parser_generate.add_argument(
        "-n", "--no-library", help="omit generating library files", action="store_true"
    )
    parser_generate.add_argument(
        "-d", "--directory", help="output directory", default=".", type=Path
    )
    parser_generate.add_argument(
        "--debug",
        help="enable adding of debug output to generated code",
        action="store_true",
    )
    parser_generate.add_argument(
        "--ignore-unsupported-checksum",
        help="ignore checksum aspects during code generation",
        action="store_true",
    )
    parser_generate.add_argument(
        "files", metavar="FILE", type=Path, nargs="*", help="specification file"
    )
    parser_generate.set_defaults(func=generate)

    parser_graph = subparsers.add_parser("graph", help="generate graphs")
    parser_graph.add_argument(
        "-f",
        "--format",
        type=str,
        default="svg",
        choices=["dot", "jpg", "pdf", "png", "raw", "svg"],
        help=("output format (default: svg)"),
    )
    parser_graph.add_argument(
        "files", metavar="FILE", type=Path, nargs="+", help="specification file"
    )
    parser_graph.add_argument("-d", "--directory", help="output directory", default=".", type=Path)
    parser_graph.add_argument(
        "--no-verification",
        action="store_true",
        help=("skip time-consuming verification of model"),
    )
    parser_graph.set_defaults(func=graph)

    args = parser.parse_args(argv[1:])

    if args.version:
        print(version())
        return 0

    if not args.subcommand:
        parser.print_usage()
        return 2

    if args.quiet:
        logging.disable(logging.CRITICAL)

    try:
        args.func(args)
    except RecordFluxError as e:
        return f"{e}"
    except Exception:  # pylint: disable = broad-except
        return f"""
------------------------------ RecordFlux Bug ------------------------------
{version()}

Optimized: {not __debug__}

Command: {' '.join(argv)}

{traceback.format_exc()}
----------------------------------------------------------------------------

A bug was detected. Please report this issue on GitHub:

https://github.com/Componolit/RecordFlux/issues/new?labels=bug

Include the complete content of the bug box shown above and all input files
in the report."""

    return 0


def version() -> str:
    dependencies = [
        f"{r.project_name} {get_distribution(r.project_name).version}"
        for r in get_distribution("RecordFlux").requires()
        if not r.project_name.startswith("RecordFlux")
    ]
    return "\n".join(
        [
            f"RecordFlux {__version__}",
            f"RecordFlux-parser {librflxlang.version}",
            *dependencies,
        ]
    )


def check(args: argparse.Namespace) -> None:
    parse(args.files)


def generate(args: argparse.Namespace) -> None:
    # WORKAROUND: Componolit/Workarounds#28
    args.prefix = args.prefix if args.prefix != " " else ""

    if args.prefix and "" in args.prefix.split("."):
        fail(f'invalid prefix: "{args.prefix}"', Subsystem.CLI)

    if not args.directory.is_dir():
        fail(f'directory not found: "{args.directory}"', Subsystem.CLI)

    model = parse(args.files)

    generator = Generator(
        model,
        args.prefix,
        reproducible=os.environ.get("RFLX_REPRODUCIBLE") is not None,
        debug=args.debug,
        ignore_unsupported_checksum=args.ignore_unsupported_checksum,
    )
    generator.write_units(args.directory)
    if not args.no_library:
        generator.write_library_files(args.directory)
    if args.prefix == DEFAULT_PREFIX:
        generator.write_top_level_package(args.directory)


def parse(files: Sequence[Path], skip_verification: bool = False) -> Model:
    parser = Parser(skip_verification, cached=True)
    error = RecordFluxError()
    present_files = []

    for f in files:
        if not f.is_file():
            error.append(f'file not found: "{f}"', Subsystem.CLI, Severity.ERROR)
            continue

        present_files.append(Path(f))

    try:
        parser.parse(*present_files)
    except RecordFluxError as e:
        error.extend(e)

    try:
        model = parser.create_model()
    except RecordFluxError as e:
        error.extend(e)

    error.propagate()
    return model


def graph(args: argparse.Namespace) -> None:
    if not args.directory.is_dir():
        fail(f'directory not found: "{args.directory}"', Subsystem.GRAPH)

    model = parse(args.files, args.no_verification)
    locations: Dict[str, Dict[str, Dict[str, Dict[str, int]]]] = defaultdict(dict)

    for m in [*model.messages, *model.sessions]:
        assert isinstance(m, (Message, Session))
        name = m.identifier.flat
        filename = args.directory.joinpath(name).with_suffix(f".{args.format}")
        Graph(m).write(filename, fmt=args.format)

        assert m.location
        assert m.location.start
        assert m.location.end
        assert m.location.source

        source = str(m.location.source)

        locations[source][name] = {
            "start": {"line": m.location.start[0], "column": m.location.start[1]},
            "end": {"line": m.location.end[0], "column": m.location.end[1]},
        }

    filename = args.directory.joinpath("locations.json")
    with open(filename, "w") as f:
        json.dump(locations, f)



from typing import Sequence, Union

from rflx.pyrflx.error import PyRFLXError


class Bitstring:
    def __init__(self, bits: str = ""):
        if not self.valid_bitstring(bits):
            raise PyRFLXError("Bitstring does not consist of only 0 and 1")
        self._bits = bits

    def __add__(self, other: "Bitstring") -> "Bitstring":
        return Bitstring(self._bits + other._bits)

    def __iadd__(self, other: "Bitstring") -> "Bitstring":
        self._bits += other._bits
        return self

    def __getitem__(self, key: Union[int, slice]) -> "Bitstring":
        if isinstance(key, slice) and isinstance(key.stop, int) and len(self._bits) < key.stop:
            raise IndexError
        return Bitstring(self._bits[key])

    def __repr__(self) -> str:
        return f'Bitstring("{self._bits}")'

    def __str__(self) -> str:
        return self._bits

    def __int__(self) -> int:
        return int(self._bits, 2)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Bitstring):
            return NotImplemented
        return self._bits == other._bits

    def __bytes__(self) -> bytes:
        return b"".join(
            [int(self._bits[i : i + 8], 2).to_bytes(1, "big") for i in range(0, len(self._bits), 8)]
        )

    def __len__(self) -> int:
        return len(self._bits)

    @classmethod
    def from_bytes(cls, msg: bytes) -> "Bitstring":
        return cls(format(int.from_bytes(msg, "big"), f"0{len(msg) * 8}b"))

    @staticmethod
    def valid_bitstring(bitstring: str) -> bool:
        return all((bit in ["0", "1"] for bit in bitstring))

    @staticmethod
    def join(iterable: Sequence["Bitstring"]) -> "Bitstring":
        joined_bitstring = Bitstring()
        for i in iterable:
            joined_bitstring += i

        return joined_bitstring



# pylint: disable = too-many-lines

from typing import List, Mapping, Sequence, Tuple

import rflx.expression as expr
from rflx.ada import (
    FALSE,
    ID,
    TRUE,
    Add,
    And,
    AndThen,
    Assignment,
    Call,
    CallStatement,
    Case,
    CaseStatement,
    Div,
    Equal,
    Expr,
    ExpressionFunctionDeclaration,
    First,
    FormalSubprogramDeclaration,
    FunctionSpecification,
    GenericFunctionInstantiation,
    GreaterEqual,
    If,
    IfStatement,
    Indexed,
    InOutParameter,
    Length,
    Less,
    Mod,
    Mul,
    NamedAggregate,
    Number,
    ObjectDeclaration,
    Old,
    Or,
    OutParameter,
    Parameter,
    Postcondition,
    Pragma,
    PragmaStatement,
    Precondition,
    ProcedureSpecification,
    Result,
    ReturnStatement,
    Selected,
    Size,
    Slice,
    String,
    Sub,
    Subprogram,
    SubprogramBody,
    SubprogramDeclaration,
    UnitPart,
    Variable,
)
from rflx.common import unique
from rflx.const import BUILTINS_PACKAGE
from rflx.model import FINAL, INITIAL, Composite, Field, Message, Scalar, Type

from . import common, const


class ParserGenerator:
    def __init__(self, prefix: str = "") -> None:
        self.prefix = prefix

    def extract_function(self, type_identifier: ID) -> Subprogram:
        return GenericFunctionInstantiation(
            "Extract",
            FunctionSpecification(
                const.TYPES * "Extract",
                type_identifier,
                [
                    Parameter(["Buffer"], const.TYPES_BYTES),
                    Parameter(["Offset"], const.TYPES_OFFSET),
                ],
            ),
            [common.prefixed_type_identifier(type_identifier, self.prefix)],
        )

    def create_internal_functions(
        self,
        message: Message,
        scalar_fields: Mapping[Field, Type],
        composite_fields: Sequence[Field],
    ) -> UnitPart:
        def result(field: Field, message: Message) -> NamedAggregate:
            aggregate: List[Tuple[str, Expr]] = [("Fld", Variable(field.affixed_name))]
            if isinstance(message.types[field], Scalar):
                aggregate.append(
                    (
                        f"{field.name}_Value",
                        Call(
                            "Extract",
                            [
                                Slice(
                                    Variable("Ctx.Buffer.all"),
                                    Variable("Buffer_First"),
                                    Variable("Buffer_Last"),
                                ),
                                Variable("Offset"),
                            ],
                        ),
                    )
                )
            elif isinstance(message.types[field], Composite) and common.is_compared_to_aggregate(
                field, message
            ):
                aggregate.append(
                    (
                        f"{field.name}_Value",
                        Slice(
                            Variable("Ctx.Buffer.all"),
                            Variable("Buffer_First"),
                            Variable("Buffer_Last"),
                        ),
                    )
                )
            return NamedAggregate(*aggregate)

        comparison_to_aggregate = any(
            (isinstance(t, Composite) and common.is_compared_to_aggregate(f, message))
            for f, t in message.types.items()
        )

        return UnitPart(
            [],
            [
                *(
                    [
                        ExpressionFunctionDeclaration(
                            FunctionSpecification(
                                "Composite_Field", "Boolean", [Parameter(["Fld"], "Field")]
                            ),
                            Case(
                                Variable("Fld"),
                                [
                                    (
                                        Variable(f.affixed_name),
                                        TRUE if f in composite_fields else FALSE,
                                    )
                                    for f in message.fields
                                ],
                            ),
                        )
                    ]
                    if scalar_fields and composite_fields
                    else []
                ),
                SubprogramBody(
                    FunctionSpecification(
                        "Get_Field_Value",
                        "Field_Dependent_Value",
                        [Parameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")],
                    ),
                    [
                        *common.field_bit_location_declarations(Variable("Fld")),
                        *(
                            common.field_byte_location_declarations()
                            if scalar_fields
                            else common.field_byte_bounds_declarations()
                        ),
                        *unique(
                            self.extract_function(common.full_base_type_name(t))
                            for f, t in message.types.items()
                            if isinstance(t, Scalar)
                        ),
                    ]
                    if scalar_fields or comparison_to_aggregate
                    else [],
                    [
                        ReturnStatement(
                            Case(
                                Variable("Fld"),
                                [
                                    (Variable(f.affixed_name), result(f, message))
                                    for f in message.fields
                                ],
                            )
                        )
                    ],
                    [
                        Precondition(
                            AndThen(
                                Call("Has_Buffer", [Variable("Ctx")]),
                                Call("Valid_Next", [Variable("Ctx"), Variable("Fld")]),
                                Call(
                                    "Sufficient_Buffer_Length", [Variable("Ctx"), Variable("Fld")]
                                ),
                            )
                        ),
                        Postcondition(
                            Equal(
                                Selected(Result("Get_Field_Value"), "Fld"),
                                Variable("Fld"),
                            )
                        ),
                    ],
                ),
            ],
        )

    @staticmethod
    def create_verify_procedure(
        message: Message,
        scalar_fields: Mapping[Field, Scalar],
        composite_fields: Sequence[Field],
    ) -> UnitPart:
        specification = ProcedureSpecification(
            "Verify", [InOutParameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")]
        )

        valid_field_condition = And(
            Call(
                "Valid_Value",
                [Variable("Value")],
            ),
            Call(
                "Field_Condition",
                [
                    Variable("Ctx"),
                    Variable("Value"),
                    *(
                        [Call("Field_Size", [Variable("Ctx"), Variable("Fld")])]
                        if common.size_dependent_condition(message)
                        else []
                    ),
                ],
            ),
        )

        set_message_last = Assignment(
            Variable("Ctx.Message_Last"),
            Mul(
                Div(
                    Add(
                        Call("Field_Last", [Variable("Ctx"), Variable("Fld")]),
                        Number(7),
                    ),
                    Number(8),
                ),
                Number(8),
            ),
        )

        set_cursors_statements = [
            *(
                [
                    PragmaStatement(
                        "Assert",
                        [
                            If(
                                [
                                    (
                                        Or(
                                            *[
                                                Equal(Variable("Fld"), Variable(f.affixed_name))
                                                for f in message.direct_predecessors(FINAL)
                                            ]
                                        ),
                                        Equal(
                                            Mod(
                                                Call(
                                                    "Field_Last", [Variable("Ctx"), Variable("Fld")]
                                                ),
                                                Size(const.TYPES_BYTE),
                                            ),
                                            Number(0),
                                        ),
                                    )
                                ]
                            )
                        ],
                    )
                ]
                if len(message.fields) > 1
                else []
            ),
            # Componolit/RecordFlux#664:
            # The provability of the context predicate is increased by duplicating the statement
            # inside a case statement.
            CaseStatement(
                Variable("Fld"),
                [
                    (
                        Variable(f.affixed_name),
                        [set_message_last],
                    )
                    for f in message.fields
                ],
                case_grouping=False,
            )
            if len(message.fields) > 1
            else set_message_last,
            IfStatement(
                [
                    (
                        Call("Composite_Field", [Variable("Fld")]),
                        [set_context_cursor_composite_field("Fld")],
                    )
                ],
                [set_context_cursor_scalar()],
            )
            if scalar_fields and composite_fields
            else set_context_cursor_scalar()
            if scalar_fields and not composite_fields
            else set_context_cursor_composite_field("Fld"),
            *(
                [
                    # ISSUE: Componolit/RecordFlux#664
                    # The provability of the context predicate is increased by splitting the
                    # assignment into multiple statements.
                    IfStatement(
                        [
                            (
                                Equal(Variable("Fld"), Variable(f.affixed_name)),
                                [
                                    Assignment(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Call("Successor", [Variable("Ctx"), Variable("Fld")]),
                                        ),
                                        NamedAggregate(
                                            ("State", Variable("S_Invalid")),
                                            ("Predecessor", Variable("Fld")),
                                        ),
                                    )
                                ],
                            )
                            for f in message.fields
                        ]
                    )
                ]
                if len(message.fields) > 1
                else []
            ),
        ]

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [
                        Postcondition(
                            And(
                                Equal(
                                    Call("Has_Buffer", [Variable("Ctx")]),
                                    Old(Call("Has_Buffer", [Variable("Ctx")])),
                                ),
                                *const.CONTEXT_INVARIANT,
                            )
                        ),
                    ],
                )
            ],
            [
                SubprogramBody(
                    specification,
                    [ObjectDeclaration(["Value"], "Field_Dependent_Value")],
                    [
                        IfStatement(
                            [
                                (
                                    AndThen(
                                        Call("Has_Buffer", [Variable("Ctx")]),
                                        Call(
                                            "Invalid",
                                            [Indexed(Variable("Ctx.Cursors"), Variable("Fld"))],
                                        ),
                                        Call(
                                            "Valid_Predecessor", [Variable("Ctx"), Variable("Fld")]
                                        ),
                                        Call("Path_Condition", [Variable("Ctx"), Variable("Fld")]),
                                    ),
                                    [
                                        IfStatement(
                                            [
                                                (
                                                    Call(
                                                        "Sufficient_Buffer_Length",
                                                        [Variable("Ctx"), Variable("Fld")],
                                                    ),
                                                    [
                                                        Assignment(
                                                            "Value",
                                                            Call(
                                                                "Get_Field_Value",
                                                                [Variable("Ctx"), Variable("Fld")],
                                                            ),
                                                        ),
                                                        IfStatement(
                                                            [
                                                                (
                                                                    valid_field_condition,
                                                                    set_cursors_statements,
                                                                )
                                                            ],
                                                            [
                                                                Assignment(
                                                                    Indexed(
                                                                        Variable("Ctx.Cursors"),
                                                                        Variable("Fld"),
                                                                    ),
                                                                    NamedAggregate(
                                                                        (
                                                                            "State",
                                                                            Variable("S_Invalid"),
                                                                        ),
                                                                        (
                                                                            "Predecessor",
                                                                            Variable(
                                                                                FINAL.affixed_name,
                                                                            ),
                                                                        ),
                                                                    ),
                                                                )
                                                            ],
                                                        ),
                                                    ],
                                                )
                                            ],
                                            [
                                                Assignment(
                                                    Indexed(
                                                        Variable("Ctx.Cursors"), Variable("Fld")
                                                    ),
                                                    NamedAggregate(
                                                        ("State", Variable("S_Incomplete")),
                                                        (
                                                            "Predecessor",
                                                            Variable(FINAL.affixed_name),
                                                        ),
                                                    ),
                                                )
                                            ],
                                        )
                                    ],
                                )
                            ],
                        )
                    ],
                )
            ],
        )

    @staticmethod
    def create_verify_message_procedure(message: Message) -> UnitPart:
        specification = ProcedureSpecification(
            "Verify_Message", [InOutParameter(["Ctx"], "Context")]
        )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [
                        Postcondition(
                            And(
                                Equal(
                                    Call("Has_Buffer", [Variable("Ctx")]),
                                    Old(Call("Has_Buffer", [Variable("Ctx")])),
                                ),
                                *const.CONTEXT_INVARIANT,
                            )
                        ),
                    ],
                )
            ],
            [
                SubprogramBody(
                    specification,
                    [],
                    [
                        CallStatement("Verify", [Variable("Ctx"), Variable(f.affixed_name)])
                        for f in message.fields
                    ],
                )
            ],
        )

    @staticmethod
    def create_present_function() -> UnitPart:
        specification = FunctionSpecification(
            "Present", "Boolean", [Parameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")]
        )

        return UnitPart(
            [SubprogramDeclaration(specification)],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    AndThen(
                        Call(
                            "Structural_Valid", [Indexed(Variable("Ctx.Cursors"), Variable("Fld"))]
                        ),
                        Less(
                            Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "First"),
                            Add(
                                Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "Last"),
                                Number(1),
                            ),
                        ),
                    ),
                )
            ],
        )

    @staticmethod
    def create_structural_valid_function() -> UnitPart:
        specification = FunctionSpecification(
            "Structural_Valid",
            "Boolean",
            [Parameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")],
        )

        return UnitPart(
            [SubprogramDeclaration(specification)],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    And(
                        Or(
                            *[
                                Equal(
                                    Selected(
                                        Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "State"
                                    ),
                                    Variable(s),
                                )
                                for s in ("S_Valid", "S_Structural_Valid")
                            ]
                        )
                    ),
                )
            ],
        )

    @staticmethod
    def create_valid_function() -> UnitPart:
        specification = FunctionSpecification(
            "Valid", "Boolean", [Parameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")]
        )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [
                        Postcondition(
                            If(
                                [
                                    (
                                        Result("Valid"),
                                        And(
                                            Call(
                                                "Structural_Valid",
                                                [Variable("Ctx"), Variable("Fld")],
                                            ),
                                            Call("Present", [Variable("Ctx"), Variable("Fld")]),
                                        ),
                                    )
                                ]
                            )
                        ),
                    ],
                )
            ],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    AndThen(
                        Equal(
                            Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "State"),
                            Variable("S_Valid"),
                        ),
                        Less(
                            Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "First"),
                            Add(
                                Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "Last"),
                                Number(1),
                            ),
                        ),
                    ),
                )
            ],
        )

    @staticmethod
    def create_incomplete_function() -> UnitPart:
        specification = FunctionSpecification(
            "Incomplete", "Boolean", [Parameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")]
        )

        return UnitPart(
            [SubprogramDeclaration(specification)],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    Equal(
                        Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "State"),
                        Variable("S_Incomplete"),
                    ),
                )
            ],
        )

    @staticmethod
    def create_invalid_function() -> UnitPart:
        specification = FunctionSpecification(
            "Invalid", "Boolean", [Parameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")]
        )

        return UnitPart(
            [SubprogramDeclaration(specification)],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    Or(
                        Equal(
                            Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "State"),
                            Variable("S_Invalid"),
                        ),
                        Equal(
                            Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "State"),
                            Variable("S_Incomplete"),
                        ),
                    ),
                )
            ],
        )

    def create_structural_valid_message_function(self, message: Message) -> UnitPart:
        specification = FunctionSpecification(
            "Structural_Valid_Message", "Boolean", [Parameter(["Ctx"], "Context")]
        )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [Precondition(Call("Has_Buffer", [Variable("Ctx")]))],
                )
            ],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    self.valid_message_condition(message, structural=True),
                )
            ],
        )

    def create_valid_message_function(self, message: Message) -> UnitPart:
        specification = FunctionSpecification(
            "Valid_Message", "Boolean", [Parameter(["Ctx"], "Context")]
        )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [Precondition(Call("Has_Buffer", [Variable("Ctx")]))],
                )
            ],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    self.valid_message_condition(message),
                )
            ],
        )

    @staticmethod
    def create_incomplete_message_function(message: Message) -> UnitPart:
        specification = FunctionSpecification(
            "Incomplete_Message", "Boolean", [Parameter(["Ctx"], "Context")]
        )

        return UnitPart(
            [SubprogramDeclaration(specification)],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    Or(
                        *[
                            Call(
                                "Incomplete",
                                [Variable("Ctx"), Variable(f.affixed_name)],
                            )
                            for f in message.fields
                        ]
                    ),
                )
            ],
        )

    def create_scalar_getter_functions(self, scalar_fields: Mapping[Field, Scalar]) -> UnitPart:
        def specification(field: Field, field_type: Type) -> FunctionSpecification:
            if field_type.package == BUILTINS_PACKAGE:
                type_identifier = ID(field_type.name)
            else:
                type_identifier = self.prefix * ID(field_type.identifier)

            return FunctionSpecification(
                f"Get_{field.name}", type_identifier, [Parameter(["Ctx"], "Context")]
            )

        def result(field: Field) -> Expr:
            return Call(
                "To_Actual",
                [
                    Selected(
                        Indexed(Variable("Ctx.Cursors"), Variable(field.affixed_name)),
                        f"Value.{field.name}_Value",
                    )
                ],
            )

        return UnitPart(
            [
                # WORKAROUND: Componolit/Workarounds#31
                Pragma("Warnings", [Variable("Off"), String("precondition is always False")]),
                *[
                    SubprogramDeclaration(
                        specification(f, t),
                        [
                            Precondition(
                                Call("Valid", [Variable("Ctx"), Variable(f.affixed_name)]),
                            )
                        ],
                    )
                    for f, t in scalar_fields.items()
                ],
                Pragma("Warnings", [Variable("On"), String("precondition is always False")]),
            ],
            private=[
                ExpressionFunctionDeclaration(specification(f, t), result(f))
                for f, t in scalar_fields.items()
            ],
        )

    @staticmethod
    def create_opaque_getter_functions(opaque_fields: Sequence[Field]) -> UnitPart:
        def name(field: Field) -> str:
            return f"Get_{field.name}"

        def specification(field: Field) -> FunctionSpecification:
            return FunctionSpecification(
                name(field),
                const.TYPES_BYTES,
                [Parameter(["Ctx"], "Context")],
            )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f),
                    [
                        Precondition(
                            AndThen(
                                Call("Has_Buffer", [Variable("Ctx")]),
                                Call(
                                    "Structural_Valid",
                                    [Variable("Ctx"), Variable(f.affixed_name)],
                                ),
                                Call(
                                    "Valid_Next",
                                    [Variable("Ctx"), Variable(f.affixed_name)],
                                ),
                            )
                        ),
                        Postcondition(
                            Equal(
                                Length(Result(name(f))),
                                Call(
                                    const.TYPES_TO_LENGTH,
                                    [
                                        Call(
                                            "Field_Size",
                                            [
                                                Variable("Ctx"),
                                                Variable(f.affixed_name),
                                            ],
                                        )
                                    ],
                                ),
                            )
                        ),
                    ],
                )
                for f in opaque_fields
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        ObjectDeclaration(
                            ["First"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "First",
                                    )
                                ],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(
                            ["Last"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "Last",
                                    )
                                ],
                            ),
                            constant=True,
                        ),
                    ],
                    [
                        ReturnStatement(
                            Slice(Variable("Ctx.Buffer.all"), Variable("First"), Variable("Last")),
                        )
                    ],
                )
                for f in opaque_fields
            ],
        )

    @staticmethod
    def create_opaque_getter_procedures(opaque_fields: Sequence[Field]) -> UnitPart:
        def specification(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Get_{field.name}",
                [Parameter(["Ctx"], "Context"), OutParameter(["Data"], const.TYPES_BYTES)],
            )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f),
                    [
                        Precondition(
                            AndThen(
                                Call("Has_Buffer", [Variable("Ctx")]),
                                Call(
                                    "Structural_Valid",
                                    [Variable("Ctx"), Variable(f.affixed_name)],
                                ),
                                Call(
                                    "Valid_Next",
                                    [Variable("Ctx"), Variable(f.affixed_name)],
                                ),
                                GreaterEqual(
                                    Length("Data"),
                                    Call(
                                        const.TYPES_TO_LENGTH,
                                        [
                                            Call(
                                                "Field_Size",
                                                [
                                                    Variable("Ctx"),
                                                    Variable(f.affixed_name),
                                                ],
                                            )
                                        ],
                                    ),
                                ),
                            )
                        )
                    ],
                )
                for f in opaque_fields
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        ObjectDeclaration(
                            ["First"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "First",
                                    )
                                ],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(
                            ["Last"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "Last",
                                    )
                                ],
                            ),
                            constant=True,
                        ),
                    ],
                    [
                        Assignment(
                            "Data",
                            NamedAggregate(("others", First(const.TYPES_BYTE))),
                        ),
                        Assignment(
                            Slice(
                                Variable("Data"),
                                First("Data"),
                                Add(First("Data"), Sub(Variable("Last"), Variable("First"))),
                            ),
                            Slice(
                                Variable("Ctx.Buffer.all"),
                                Variable("First"),
                                Variable("Last"),
                            ),
                        ),
                    ],
                )
                for f in opaque_fields
            ],
        )

    @staticmethod
    def create_generic_opaque_getter_procedures(opaque_fields: Sequence[Field]) -> UnitPart:
        def specification(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Generic_Get_{field.name}", [Parameter(["Ctx"], "Context")]
            )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f),
                    [
                        Precondition(
                            And(
                                Call("Has_Buffer", [Variable("Ctx")]),
                                Call(
                                    "Present",
                                    [Variable("Ctx"), Variable(f.affixed_name)],
                                ),
                            )
                        )
                    ],
                    [
                        FormalSubprogramDeclaration(
                            ProcedureSpecification(
                                f"Process_{f.name}", [Parameter([f.name], const.TYPES_BYTES)]
                            )
                        )
                    ],
                )
                for f in opaque_fields
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        ObjectDeclaration(
                            ["First"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "First",
                                    )
                                ],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(
                            ["Last"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "Last",
                                    )
                                ],
                            ),
                            constant=True,
                        ),
                    ],
                    [
                        CallStatement(
                            f"Process_{f.name}",
                            [
                                Slice(
                                    Variable("Ctx.Buffer.all"), Variable("First"), Variable("Last")
                                )
                            ],
                        )
                    ],
                )
                for f in opaque_fields
            ],
        )

    def valid_message_condition(
        self, message: Message, field: Field = INITIAL, structural: bool = False
    ) -> Expr:
        def condition(message: Message, field: Field, structural: bool) -> expr.Expr:
            return expr.Or(
                *[
                    l.condition
                    if l.target == FINAL
                    else expr.AndThen(
                        expr.Call(
                            "Structural_Valid"
                            if structural and isinstance(message.types[l.target], Composite)
                            else "Valid",
                            [
                                expr.Variable("Ctx"),
                                expr.Variable(l.target.affixed_name, immutable=True),
                            ],
                        ),
                        l.condition,
                        condition(message, l.target, structural),
                    )
                    for l in message.outgoing(field)
                ]
            )

        return (
            condition(message, field, structural)
            .substituted(common.substitution(message, self.prefix))
            .simplified()
            .ada_expr()
        )


def set_context_cursor_scalar() -> Assignment:
    return Assignment(
        Indexed(Variable("Ctx.Cursors"), Variable("Fld")),
        NamedAggregate(
            ("State", Variable("S_Valid")),
            ("First", Call("Field_First", [Variable("Ctx"), Variable("Fld")])),
            ("Last", Call("Field_Last", [Variable("Ctx"), Variable("Fld")])),
            ("Value", Variable("Value")),
            (
                "Predecessor",
                Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "Predecessor"),
            ),
        ),
    )


def set_context_cursor_composite_field(field_name: str) -> Assignment:
    return Assignment(
        Indexed(
            Variable("Ctx.Cursors"),
            Variable(field_name),
        ),
        NamedAggregate(
            ("State", Variable("S_Structural_Valid")),
            (
                "First",
                Call(
                    "Field_First",
                    [Variable("Ctx"), Variable(field_name)],
                ),
            ),
            (
                "Last",
                Call(
                    "Field_Last",
                    [Variable("Ctx"), Variable(field_name)],
                ),
            ),
            ("Value", Variable("Value")),
            (
                "Predecessor",
                Selected(
                    Indexed(
                        Variable("Ctx.Cursors"),
                        Variable(field_name),
                    ),
                    "Predecessor",
                ),
            ),
        ),
    )



import re

from setuptools import find_packages, setup  # type: ignore

with open("rflx/__init__.py") as f:
    match = re.search(r'__version__ = "(.*?)"', f.read())
    assert match
    version = match.group(1)

with open("README.md") as f:
    readme = f.read()

setup(
    name="RecordFlux",
    version=version,
    description=(
        "A toolset for the formal specification of messages and the generation of "
        "verifiable binary parsers and message generators."
    ),
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Tobias Reiher",
    author_email="reiher@componolit.com",
    url="https://github.com/Componolit/RecordFlux",
    license="AGPL-3.0",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Ada",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Communications",
        "Topic :: Security",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Code Generators",
        "Topic :: System :: Networking",
    ],
    packages=find_packages(include=("rflx", "rflx.*")),
    package_data={"rflx": ["py.typed", "templates/*"]},
    python_requires=">=3.7",
    install_requires=[
        "attrs >=20, <22",
        "icontract >=2.3.4, <3",
        "pydotplus >=2, <3",
        "z3-solver >=4, <5",
        "RecordFlux-language@git+https://github.com/Componolit/RecordFlux-language.git"
        "@v0.6.0#egg=RecordFlux-language",
    ],
    extras_require={
        "devel": [
            "black ==20.8b1",
            "flake8 >=3.9.2, <3.10",
            "hypothesis >=6.14, <6.15",
            "isort >=5.9.1, <5.10",
            "mypy ==0.910",
            "pyicontract-lint >=2.1.2, <2.2",
            "pylint >=2.8.3, <2.9",
            "pytest >=6.2.4, <6.3",
            "pytest-cov >=2.12.1, <2.13",
            "pytest-xdist >=2.3.0, <2.4",
            "tqdm >=4.61.1, <4.62",
            "types-pkg_resources >=0.1.3, <0.2",
        ]
    },
    scripts=["bin/rflx"],
)



from typing import Dict, Sequence

from rflx import const
from rflx.common import Base, verbose_repr
from rflx.error import RecordFluxError, Severity, Subsystem
from rflx.identifier import ID

from . import message, session, type_


class Model(Base):
    def __init__(
        self, types: Sequence[type_.Type] = None, sessions: Sequence[session.Session] = None
    ) -> None:
        self.__types = types or []
        self.__sessions = sessions or []

        self.__validate()

    def __repr__(self) -> str:
        return verbose_repr(self, ["types"])

    def __str__(self) -> str:
        return "\n\n".join(
            f"{t};"
            for t in self.__types
            if not type_.is_builtin_type(t.name) and not type_.is_internal_type(t.name)
        )

    @property
    def types(self) -> Sequence[type_.Type]:
        return self.__types

    @property
    def messages(self) -> Sequence[message.Message]:
        return [m for m in self.__types if isinstance(m, message.Message)]

    @property
    def refinements(self) -> Sequence[message.Refinement]:
        return [m for m in self.__types if isinstance(m, message.Refinement)]

    @property
    def sessions(self) -> Sequence[session.Session]:
        return self.__sessions

    def __validate(self) -> None:
        error = self.__check_duplicates()
        error += self.__check_conflicts()
        error.propagate()

    def __check_duplicates(self) -> RecordFluxError:
        error = RecordFluxError()
        types: Dict[ID, type_.Type] = {}
        sessions: Dict[ID, session.Session] = {}

        for t in self.__types:
            if t.identifier in types:
                error.append(
                    f'conflicting refinement of "{t.pdu.identifier}" with "{t.sdu.identifier}"'
                    if isinstance(t, message.Refinement)
                    else f'name conflict for type "{t.identifier}"',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    t.location,
                )
                error.append(
                    "previous occurrence of refinement"
                    if isinstance(t, message.Refinement)
                    else f'previous occurrence of "{t.identifier}"',
                    Subsystem.MODEL,
                    Severity.INFO,
                    types[t.identifier].location,
                )
            types[t.identifier] = t

        for s in self.__sessions:
            if s.identifier in types or s.identifier in sessions:
                error.append(
                    f'name conflict for session "{s.identifier}"',
                    Subsystem.MODEL,
                    Severity.ERROR,
                    s.location,
                )
                error.append(
                    f'previous occurrence of "{s.identifier}"',
                    Subsystem.MODEL,
                    Severity.INFO,
                    types[s.identifier].location
                    if s.identifier in types
                    else sessions[s.identifier].location,
                )
            sessions[s.identifier] = s

        return error

    def __check_conflicts(self) -> RecordFluxError:
        error = RecordFluxError()

        for e1, e2 in [
            (e1, e2)
            for i1, e1 in enumerate(self.__types)
            for i2, e2 in enumerate(self.__types)
            if (
                isinstance(e1, type_.Enumeration)
                and isinstance(e2, type_.Enumeration)
                and i1 < i2
                and (
                    e1.package == e2.package
                    or e1.package == const.BUILTINS_PACKAGE
                    or e2.package == const.BUILTINS_PACKAGE
                )
            )
        ]:
            identical_literals = set(e2.literals) & set(e1.literals)

            if identical_literals:
                literals_message = ", ".join([f"{l}" for l in sorted(identical_literals)])
                error.append(
                    f"conflicting literals: {literals_message}",
                    Subsystem.MODEL,
                    Severity.ERROR,
                    e2.location,
                )
                error.extend(
                    [
                        (
                            f'previous occurrence of "{l}"',
                            Subsystem.MODEL,
                            Severity.INFO,
                            l.location,
                        )
                        for l in sorted(identical_literals)
                    ]
                )

        literals = [
            t.package * l
            for t in self.__types
            if isinstance(t, type_.Enumeration)
            for l in t.literals
        ]
        name_conflicts = [
            (l, t)
            for l in literals
            for t in self.__types
            if (l.parent == t.package or type_.is_builtin_type(t.identifier))
            and l.name == t.identifier.name
        ]
        for literal, conflicting_type in name_conflicts:
            error.append(
                f'literal "{literal.name}" conflicts with type declaration',
                Subsystem.MODEL,
                Severity.ERROR,
                literal.location,
            )
            error.append(
                f'conflicting type "{conflicting_type.identifier}"',
                Subsystem.MODEL,
                Severity.INFO,
                conflicting_type.location,
            )

        return error



import re
from abc import ABC
from typing import Iterable, Iterator, Sequence, Set, TypeVar


class Base(ABC):
    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            for k in other.__dict__:
                if k not in ["location", "error"] and k not in self.__dict__:
                    return False
            for k, v in self.__dict__.items():
                if k not in ["location", "error"] and (
                    k not in other.__dict__ or v != other.__dict__[k]
                ):
                    return False
            return True
        return NotImplemented

    def __repr__(self) -> str:
        args = "\n" + ",\n".join(f"{k}={v!r}" for k, v in self.__dict__.items() if k != "location")
        return indent_next(f"\n{self.__class__.__name__}({indent(args, 4)})", 4)


def verbose_repr(obj: object, attributes: Sequence[str]) -> str:
    def prefixed_str(obj: object) -> str:
        obj_str = str(obj)
        return (
            "\n" + "\n".join(f"# {l}" if l else "#" for l in obj_str.split("\n")) + "\n"
            if obj_str
            else ""
        )

    indentation = len(obj.__class__.__name__) + 1
    args = "".join(f"{a}={getattr(obj, a)!r},\n" for a in attributes)
    return indent_next(
        f"\n{obj.__class__.__name__}({indent_next(args, indentation)}\n){prefixed_str(obj)}", 4
    )


def indent(string: str, indentation: int) -> str:
    return "\n".join((indentation * " " + l if l else "") for l in string.split("\n"))


def indent_next(string: str, indentation: int) -> str:
    return string.replace("\n", "\n" + indentation * " ")


def file_name(name: str) -> str:
    return re.sub(r"(?:\.|::)", "-", name.lower())


T = TypeVar("T")  # pylint: disable=invalid-name


def unique(iterable: Iterable[T]) -> Iterator[T]:
    seen: Set[T] = set()
    for e in iterable:
        if e not in seen:
            seen.add(e)
            yield e



RESERVED_WORDS = [
    # Ada
    "abort",
    "abs",
    "abstract",
    "accept",
    "access",
    "aliased",
    "all",
    "and",
    "array",
    "at",
    "begin",
    "body",
    "case",
    "constant",
    "declare",
    "delay",
    "delta",
    "digits",
    "do",
    "else",
    "elsif",
    "end",
    "entry",
    "exception",
    "exit",
    "for",
    "function",
    "generic",
    "goto",
    "if",
    "in",
    "interface",
    "is",
    "limited",
    "loop",
    "mod",
    "new",
    "not",
    "null",
    "of",
    "or",
    "others",
    "out",
    "overriding",
    "package",
    "pragma",
    "private",
    "procedure",
    "protected",
    "raise",
    "range",
    "record",
    "rem",
    "renames",
    "requeue",
    "return",
    "reverse",
    "select",
    "separate",
    "some",
    "subtype",
    "synchronized",
    "tagged",
    "task",
    "terminate",
    "then",
    "type",
    "until",
    "use",
    "when",
    "while",
    "with",
    "xor",
    # Code generation
    "initial",
    "final",
    "ctx",
    "val",
    "enum",
]
